from argparse import ArgumentParser
from collections import defaultdict
from itertools import product
from pathlib import Path
import typing as T

from pandas import DataFrame
import yaml

from parser import parse_mother_request
from grib import read_grib_params


PORTFOLIO_PATH = Path(__file__).resolve().parent.parent / "data_portfolio"
PORTFOLIO_KINDS = ["production", "reduced"]
PORTFOLIO_VERSIONS = [path.name for path in PORTFOLIO_PATH.iterdir()]
GENERATIONS = {
    1: "v0.1.1",
    2: "v1.2.0",
}
OUTPUT_PATH = Path(__file__).resolve().parent.parent / "variables"


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        help="Path for mother request",
        type=str,
    )
    return parser


def get_variables(path: Path | str, must_match: bool):
    df = parse_mother_request(path)
    df = df[["Application", "Request No.", "gsv_param", "opa_variable"]]
    variables = defaultdict(list)
    for _, row in df.iterrows():
        gsv = row["gsv_param"]
        opa = row["opa_variable"]
        app = row["Application"]
        if gsv == opa:
            variables[gsv].append(app)
        else:
            if must_match:
                raise ValueError(
                    "GSV and OPA request have non-matching variables: "
                    f"'{row['gsv_param']}' and '{row['opa_variable']}'."
                )
            variables[gsv].append(app)
            variables[opa].append(app)
    return df, dict(variables)


def read_data_portfolio(version: str, kind: str):
    if version not in PORTFOLIO_VERSIONS:
        raise ValueError(
            f"Version '{version}' is not valid. "
            f"Please, pass one of '{','.join(PORTFOLIO_VERSIONS)}'."
        )
    if kind not in PORTFOLIO_KINDS:
        raise ValueError(
            f"Kind '{kind}' is not valid. "
            f"Please, pass one of '{','.join(PORTFOLIO_KINDS)}'."
        )
    portfolio_path = PORTFOLIO_PATH / version / kind / "portfolio.yaml"
    if not portfolio_path.exists():
        raise FileNotFoundError(f"Malformed portfolio: {version}.")
    with open(portfolio_path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)


def get_portfolios():
    portfolios = {}
    for kind, (generation, version) in product(PORTFOLIO_KINDS, GENERATIONS.items()):
        portfolios[(kind, generation)] = (
            read_data_portfolio(version=version, kind=kind)
        )
    return portfolios


def find_matches(grib: DataFrame, variables: T.Sequence[str]):
    return {
        variable: grib[grib["Short name"] == variable]
        for variable
        in variables
    }


def _record_to_dct(record, models=None):
    variable_dct = (
        record.to_dict(orient="records")[0]
    )
    format_entry = variable_dct.pop("Format", None)
    if format_entry:
        variable_dct.update({
            "GRIB1": "GRIB1" in format_entry,
            "GRIB2": "GRIB2" in format_entry,
        })
    if models:
        variable_dct.update({
            "IFS-FESOM": "ifs-fesom" in models,
            "IFS-NEMO": "ifs-nemo" in models,
            "ICON": "icon" in models,
        })
    return variable_dct

def convert_variables(portfolio, grib: DataFrame):
    variables = {}
    variable_models = defaultdict(set)
    for model, model_entries in portfolio.items():
        for entry in model_entries:
            for variable in entry["variables"]:
                variable_models[variable].add(model)
    for model, model_entries in portfolio.items():
        for entry in model_entries:
            for variable in entry["variables"]:
                if variable not in variables:
                    variable_record = grib[grib["ID"] == variable]
                    variables[variable] = _record_to_dct(
                        variable_record, variable_models[variable]
                    )
    return DataFrame.from_records(list(variables.values()))


def dump_variables(portfolios, grib):
    for key, portfolio in portfolios.items():
        file_name = f"gen{key[1]}_{key[0]}.csv"
        variables_df = convert_variables(portfolio, grib)
        variables_df.to_csv(OUTPUT_PATH / file_name)


def main():
    args = _get_parser().parse_args()
    path = Path(args.path)
    apps, variables_apps = get_variables(path, must_match=True)
    portfolios = get_portfolios()
    grib = read_grib_params()
    dump_variables(portfolios, grib)


if __name__ == "__main__":
    main()
